import bpy
import socket
from . import properties

def is_valid(act: bpy.types.Action) -> bool:
    if act is None:
        return False
    try:
        _ = act.users
    except ReferenceError:
        return False
    return True

def is_valid(obj: bpy.types.Object) -> bool:
    if obj is None:
        return False
    try:
        _ = obj.name
    except ReferenceError:
        return False
    return True

def is_valid(prop: bpy.types.PropertyGroup) -> bool:
    if prop is None:
        return False
    try:
        _ = prop.name
    except ReferenceError:
        return False
    return True
        
def find_prop(id: int) -> properties.MocopiAvatarProperty:
    list = bpy.context.scene.mocopi_property.avatars
    if list is None:
        return None
    for prop in list:
        if prop.id == id:
            return prop
    return None

def get_ip_address():
        ip_address = '127.0.0.1'
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
                s.connect(("8.8.8.8", 80))
                ip_address = s.getsockname()[0]
        except Exception as e:
            pass
        return ip_address
        # return '192.168.1.1'

def get_scale_value(obj: bpy.types.Object) -> float:
    if not is_valid(obj):
        return 1.0
    return (obj.scale.x + obj.scale.y + obj.scale.z) / 3.0

def get_bone_name(bone: bpy.types.PoseBone) -> str:
    try:
        if bone and isinstance(bone.name, bytes):
            return bone.name.decode('utf-8')
        return bone.name if bone else ''
    except (UnicodeDecodeError, AttributeError):
        return ''