import bpy
from abc import ABCMeta, abstractmethod
from . import main
from . import models
from . import utils
from . import strings

class IAvatarPanelListener(metaclass=ABCMeta):

    @abstractmethod
    def on_drawed(self, panel, context: bpy.types.Context, id: int) -> models.Avatar:
        pass

    @abstractmethod
    def on_remove_button_clicked(self, op, context: bpy.types.Context, id: int):
        pass

    @abstractmethod
    def on_connect_button_clicked(self, op, context: bpy.types.Context, id: int):
        pass

    @abstractmethod
    def on_reboot_button_clicked(self, op, context: bpy.types.Context, id: int):
        pass

    @abstractmethod
    def on_debug_button_clicked(self, op, context: bpy.types.Context, id: int):
        pass


class MOCOPI_RECEIVER_OT_AvatarPanel(bpy.types.Operator):
    
    bl_idname = "mocopi_receiver_plugin.avatar_panel"
    bl_label = strings.get('bl_label_avatar_panel')
    bl_description = strings.get('bl_desc_avatar_panel')
    bl_options = {'INTERNAL'}

    id: bpy.props.IntProperty() # type: ignore
    action: bpy.props.StringProperty() # type: ignore
    listener: IAvatarPanelListener = None

    def execute(self, context):

        if self.action == 'remove_button':
            self.listener.on_remove_button_clicked(self, context, self.id)
        elif self.action == 'connect_button':
            self.listener.on_connect_button_clicked(self, context, self.id)
        elif self.action == 'reboot_button':
            self.listener.on_reboot_button_clicked(self, context, self.id)
        elif self.action == 'debug_button':
            self.listener.on_debug_button_clicked(self, context, self.id)

        return {'FINISHED'} 
    

class MOCOPI_RECEIVER_LIST_Enum(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        layout.label(text=item.name)


class MOCOPI_RECEIVER_PT_AvatarPanel(bpy.types.Panel):

    id: bpy.props.IntProperty(default=0) # type: ignore
    listener: IAvatarPanelListener = None
    
    # bl_idname = "MOCOPI_RECEIVER_PT_AvatarPanel"
    # bl_label = "Avatar"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "mocopi"

    def draw_header(self, context):

        layout = self.layout
        row = layout.row()
        op = row.operator(MOCOPI_RECEIVER_OT_AvatarPanel.bl_idname, text='', icon='TRASH') 
        op.id = self.id
        op.action = 'remove_button'

    def draw(self, context):

        MOCOPI_RECEIVER_OT_AvatarPanel.listener = MOCOPI_RECEIVER_PT_AvatarPanel.listener

        layout = self.layout
    
        # 更新のコールバック
        avatar = self.listener.on_drawed(self, context, self.id)
        if avatar is None:
            layout.label(text=strings.get('msg_missing_receiver'))
            # TODO: 保留
            # op = layout.operator(MOCOPI_RECEIVER_OT_AvatarPanel.bl_idname, text=strings.get('reboot_mocopi_receiver'))
            # op.id = self.id
            # op.action = 'reboot_button'
            return
        
        prop = utils.find_prop(self.id)
            
        row = layout.row()
        row.label(text=strings.get('number_port') + ':')
        row.prop(prop, 'port', text='')
        row.enabled = not avatar.running

        # TODO: 保留
        # row = layout.row()
        # row.label(text=strings.get('mirror') + ':')
        # row.prop(prop, 'is_mirror', text='', icon='MOD_MIRROR')

        row = layout.row()
        row.label(text=strings.get('connecting') if avatar.running else strings.get('unconnected'), icon='PLUGIN')
        op = row.operator(MOCOPI_RECEIVER_OT_AvatarPanel.bl_idname, text=strings.get('connect') if not avatar.running else strings.get('disconnect'))
        op.id = self.id
        op.action = 'connect_button'
        row.enabled = avatar.has_rig()

        # ターゲット
        box = layout.box() 
        row = box.row()
        row.prop_search(prop, 'rig', search_data=bpy.data, search_property='objects', text=strings.get('character'))
        # row.enabled = not avatar.running and not avatar.is_retargeted
        box.enabled = not avatar.running # not avatar.running and not avatar.is_retargeted
        
        # ボーン
        if avatar.has_rig():
            box.prop_search(prop, 'root', search_data=avatar.rig.data, search_property='bones', text='Hip')
            box.prop_search(prop, 'torso_1', search_data=avatar.rig.data, search_property='bones', text='Torso1')
            box.prop_search(prop, 'torso_2', search_data=avatar.rig.data, search_property='bones', text='Torso2')
            box.prop_search(prop, 'torso_3', search_data=avatar.rig.data, search_property='bones', text='Torso3')
            box.prop_search(prop, 'torso_4', search_data=avatar.rig.data, search_property='bones', text='Torso4')
            box.prop_search(prop, 'torso_5', search_data=avatar.rig.data, search_property='bones', text='Torso5')
            box.prop_search(prop, 'torso_6', search_data=avatar.rig.data, search_property='bones', text='Torso6')
            box.prop_search(prop, 'torso_7', search_data=avatar.rig.data, search_property='bones', text='Torso7')
            box.prop_search(prop, 'neck_1', search_data=avatar.rig.data, search_property='bones', text='Neck1')
            box.prop_search(prop, 'neck_2', search_data=avatar.rig.data, search_property='bones', text='Neck2')
            box.prop_search(prop, 'head', search_data=avatar.rig.data, search_property='bones', text='Head')
            box.prop_search(prop, 'l_shoulder', search_data=avatar.rig.data, search_property='bones', text='L_Shoulder')
            box.prop_search(prop, 'l_up_arm', search_data=avatar.rig.data, search_property='bones', text='L_UpperArm')
            box.prop_search(prop, 'l_low_arm', search_data=avatar.rig.data, search_property='bones', text='L_LowerArm')
            box.prop_search(prop, 'l_hand', search_data=avatar.rig.data, search_property='bones', text='L_Hand')
            box.prop_search(prop, 'r_shoulder', search_data=avatar.rig.data, search_property='bones', text='R_Shoulder')
            box.prop_search(prop, 'r_up_arm', search_data=avatar.rig.data, search_property='bones', text='R_UpperArm')
            box.prop_search(prop, 'r_low_arm', search_data=avatar.rig.data, search_property='bones', text='R_LowerArm')
            box.prop_search(prop, 'r_hand', search_data=avatar.rig.data, search_property='bones', text='R_Hand')
            box.prop_search(prop, 'l_up_leg', search_data=avatar.rig.data, search_property='bones', text='L_UpperLeg')
            box.prop_search(prop, 'l_low_leg', search_data=avatar.rig.data, search_property='bones', text='L_LowerLeg')
            box.prop_search(prop, 'l_foot', search_data=avatar.rig.data, search_property='bones', text='L_Foot')
            box.prop_search(prop, 'l_toe', search_data=avatar.rig.data, search_property='bones', text='L_ToeBase')
            box.prop_search(prop, 'r_up_leg', search_data=avatar.rig.data, search_property='bones', text='R_UpperLeg')
            box.prop_search(prop, 'r_low_leg', search_data=avatar.rig.data, search_property='bones', text='R_LowerLeg')
            box.prop_search(prop, 'r_foot', search_data=avatar.rig.data, search_property='bones', text='R_Foot')
            box.prop_search(prop, 'r_toe', search_data=avatar.rig.data, search_property='bones', text='R_ToeBase')


class IPanelListener(metaclass=ABCMeta):

    @abstractmethod
    def on_add_button_clicked(self, op, context: bpy.types.Context):
        pass


class MOCOPI_RECEIVER_OT_Panel(bpy.types.Operator):
    
    bl_idname = "mocopi_receiver_plugin.panel"
    bl_label = strings.get('bl_label_panel')
    bl_description = strings.get('bl_desc_panel')
    bl_options = set()

    action: bpy.props.StringProperty() # type: ignore
    listener: IPanelListener = None

    def execute(self, context):
        if self.action == 'add_button':
            self.listener.on_add_button_clicked(self, context)
        return {'FINISHED'} 
    

class MOCOPI_RECEIVER_PT_Panel(bpy.types.Panel):

    bl_idname = "MOCOPI_RECEIVER_PT_Panel"
    bl_label = strings.get('mocopi')
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "mocopi"

    listener: IPanelListener = None

    def draw_header(self, context):
        layout = self.layout
        layout.label(text="", icon_value=main.Main.image_collection["mocopi_icon"].icon_id)

    def draw(self, context):

        MOCOPI_RECEIVER_OT_Panel.listener = MOCOPI_RECEIVER_PT_Panel.listener

        layout = self.layout
        layout.label(text=strings.get('ip_address') + ': ' + utils.get_ip_address())

        op = layout.operator(MOCOPI_RECEIVER_OT_Panel.bl_idname, text=strings.get('add_mocopi_receiver'), icon='ADD')
        op.action = 'add_button'
